#lang scheme

(require "find.ss"
         "test.ss")

(provide valid-move?)

;; valid-move? : state nat nat state
;;  We assume well-formed states (e.g., no two players
;;  on the same blank space), a valid player number,
;;  and a valid roll where 0 means "no roll".
(define (valid-move? state player roll new-state)
  (let ([init (list-ref state player)]
        [dest (list-ref new-state player)])
    (and
     ;; Only the indicated player moved:
     (equal? (remq init state)
             (remq dest new-state))
     ;; Use list of possible:
     (let ([dests (find-dests state player roll)])
       (if (equal? dests #hash())
           ;; Cannot move; must stay put:
           (equal? init dest)
           ;; Destination is in possible?:
           (hash-ref dests dest #f))))))

;; Scarlet moves 1
(test (valid-move? '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))
                   0
                   1
                   '((17 23) (24 17) (15 0) (10 0) (0 6) (0 19)))
      #t)
;; Non-Scarlet mustn't move
(test (valid-move? '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))
                   0
                   1
                   '((17 23) (23 17) (15 0) (10 0) (0 6) (0 19)))
      #f)
;; Scarlet cannot return to starting spot:
(test (valid-move? '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))
                   0
                   2
                   '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19)))
      #f)
;; Scarlet reaches lounge:
(test (valid-move? '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))
                   0
                   7
                   '((U V) (24 17) (15 0) (10 0) (0 6) (0 19)))
      #t)
;; Scarlet waits outside lounge:
(test (valid-move? '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))
                   0
                   7
                   '((18 18) (24 17) (15 0) (10 0) (0 6) (0 19)))
      #t)
;; Scarlet reaches lounge with moves to spare:
(test (valid-move? '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))
                   0
                   10
                   '((U V) (24 17) (15 0) (10 0) (0 6) (0 19)))
      #t)
;; Scarlet reaches lounge, but marked incorrectly as by-suggestion:
(test (valid-move? '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))
                   0
                   7
                   '((U S) (24 17) (15 0) (10 0) (0 6) (0 19)))
      #f)
;; Scarlet moves 2, jumping over Mustard:
(test (valid-move? '((17 24) (17 23) (15 0) (10 0) (0 6) (0 19))
                   0
                   2
                   '((17 22) (17 23) (15 0) (10 0) (0 6) (0 19)))
      #t)
;; Scarlet cannot land on Mustard:
(test (valid-move? '((17 24) (17 22) (15 0) (10 0) (0 6) (0 19))
                   0
                   2
                   '((17 22) (17 22) (15 0) (10 0) (0 6) (0 19)))
      #f)
;; Scarlet cannot reach lounge if Mustard blocks the door:
(test (valid-move? '((17 24) (18 18) (15 0) (10 0) (0 6) (0 19))
                   0
                   7
                   '((U V) (18 18) (15 0) (10 0) (0 6) (0 19)))
      #f)
;; Scarlet takes secret passage:
(test (valid-move? '((U S) (18 18) (15 0) (10 0) (0 6) (0 19))
                   0
                   0
                   '((C V) (18 18) (15 0) (10 0) (0 6) (0 19)))
      #t)
;; Scarlet can't roll and then take secret passage:
(test (valid-move? '((U S) (18 18) (15 0) (10 0) (0 6) (0 19))
                   0
                   7
                   '((C V) (18 18) (15 0) (10 0) (0 6) (0 19)))
      #f)

;; QUESTIONABLE:
;; If you can't move, does staying in place count as valid?
;;  We say "yes", because we want to use this as a move checker later.
;; Scarlet cannot move at all; blocked by Mustard:
(test (valid-move? '((17 24) (17 23) (15 0) (10 0) (0 6) (0 19))
                   0
                   1
                   '((17 24) (17 23) (15 0) (10 0) (0 6) (0 19)))
      #t)
